#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""CLI de catnose. Invocable como `catnose` (console script) o `python -m catnose`."""

import argparse
import json
import sys
from pathlib import Path

from ._core import transform


def main() -> None:
  parser = argparse.ArgumentParser(prog="catnose", description="Transforma un array JSON o un objeto JSON aplicando un mapeador declarativo.", epilog="\nEjemplos:\n  %(prog)s input.json transform.json\n  %(prog)s input.json transform.json -o output.json\n")
  parser.add_argument("input", help="Fichero JSON de entrada (array de objetos o un único objeto)")
  parser.add_argument("transform_file", metavar="transform", help="Fichero JSON de transformación (mapeador)")
  parser.add_argument("-o", "--output", metavar="FILE", help="Escribe el resultado en FILE (por defecto: stdout)")
  args = parser.parse_args()

  input_path = Path(args.input)
  if not input_path.exists():
    print(f"ERROR: '{input_path}' no encontrado", file=sys.stderr)
    sys.exit(1)
  transform_path = Path(args.transform_file)
  if not transform_path.exists():
    print(f"ERROR: '{transform_path}' no encontrado", file=sys.stderr)
    sys.exit(1)

  items = json.loads(input_path.read_text(encoding="utf-8"))
  if not isinstance(items, (list, dict)):
    print("ERROR: el fichero de entrada debe ser un array JSON o un objeto JSON", file=sys.stderr)
    sys.exit(1)

  mapper = json.loads(transform_path.read_text(encoding="utf-8"))
  result = transform(items, mapper)
  output = json.dumps(result, ensure_ascii=False, indent=2)

  if args.output:
    Path(args.output).write_text(output, encoding="utf-8")
    print(f"OK: {len(result)} elementos escritos en '{args.output}'")
  else:
    print(output)


if __name__ == "__main__":
  main()



# -*- coding: utf-8 -*-
"""
catnose — Motor de transformación declarativa de JSON.

Uso básico::

    from catnose import transform

    result = transform(items, mapper)

Para usar el motor de plantillas directamente::

    from catnose import template_engine

    html = template_engine.render("Hola {{ name }}", {"name": "mundo"})
"""

from ._core import transform
from . import _template as template_engine

__all__ = ["transform", "template_engine"]
__version__ = "1.1.0"



# -*- coding: utf-8 -*-
"""
Motor de plantillas con semántica Python eval.

Sintaxis soportada (anidable):
  {{ expr }}                  — evalúa expr Python, inserta resultado como str
  {% for var in list_expr %}  — itera lista; var queda disponible en el contexto
  {% endfor %}
  {% if bool_expr %}          — bloque condicional
  {% else %}                  — rama else (opcional)
  {% endif %}

El contexto de evaluación usa DotDict para acceso por punto (x.y).
"""

import re

# ---------------------------------------------------------------------------
# Colector de errores de evaluación (opcional)
# ---------------------------------------------------------------------------
_error_collector: list | None = None
_error_seen: set = set()


def set_error_collector(lst: list | None) -> None:
  """Activa (con una lista) o desactiva (con None) la recogida de errores."""
  global _error_collector, _error_seen
  _error_collector = lst
  _error_seen = set()


def _collect_error(expr: str, exc: Exception) -> None:
  if _error_collector is not None and expr not in _error_seen:
    _error_seen.add(expr)
    _error_collector.append((expr, exc))

# ---------------------------------------------------------------------------
# Builtins seguros expuestos a eval
# ---------------------------------------------------------------------------
_BUILTINS: dict = {
  '__builtins__': {},
  'str': str, 'int': int, 'float': float, 'bool': bool,
  'len': len, 'list': list, 'dict': dict, 'tuple': tuple,
  'range': range, 'enumerate': enumerate, 'zip': zip,
  'min': min, 'max': max, 'sum': sum, 'abs': abs, 'round': round,
  'sorted': sorted, 'reversed': reversed,
  'True': True, 'False': False, 'None': None,
  'isinstance': isinstance, 'hasattr': hasattr,
  'repr': repr, 'chr': chr, 'ord': ord,
  're': re,
}

# ---------------------------------------------------------------------------
# Tokenizador
# ---------------------------------------------------------------------------
_TAG_RE = re.compile(r'\{\{(.*?)\}\}|\{%(.*?)%\}', re.DOTALL)


def _tokenize(s: str) -> list[tuple]:
  """Divide la cadena de plantilla en una lista plana de tokens."""
  s = str(s)
  tokens: list[tuple] = []
  pos = 0
  for m in _TAG_RE.finditer(s):
    if m.start() > pos:
      tokens.append(('text', s[pos:m.start()]))
    if m.group(1) is not None:
      tokens.append(('expr', m.group(1).strip()))
    else:
      inner = m.group(2).strip()
      if fm := re.match(r'^for\s+(\w+)\s+in\s+(.+)$', inner, re.DOTALL):
        tokens.append(('for', fm.group(1), fm.group(2).strip()))
      elif fm := re.match(r'^if\s+(.+)$', inner, re.DOTALL):
        tokens.append(('if', fm.group(1).strip()))
      elif inner == 'else':
        tokens.append(('else',))
      elif inner == 'endfor':
        tokens.append(('endfor',))
      elif inner == 'endif':
        tokens.append(('endif',))
      else:
        tokens.append(('text', m.group(0)))  # tag desconocido — conservar tal cual
    pos = m.end()
  if pos < len(s):
    tokens.append(('text', s[pos:]))
  return tokens

# ---------------------------------------------------------------------------
# Helpers de contexto
# ---------------------------------------------------------------------------

class _StrictDotDict(dict):
  """DotDict que lanza AttributeError en acceso a clave inexistente."""
  def __getattr__(self, name: str):
    try:
      return self[name]
    except KeyError:
      raise AttributeError(f"'{name}' no existe en el contexto") from None
  __setattr__ = dict.__setitem__
  __delattr__ = dict.__delitem__


def _to_dotdict(obj):
  """Convierte recursivamente dicts (y listas de dicts) a _StrictDotDict."""
  if isinstance(obj, dict):
    return _StrictDotDict({k: _to_dotdict(v) for k, v in obj.items()})
  if isinstance(obj, list):
    return [_to_dotdict(v) for v in obj]
  return obj


def _eval(expr: str, ctx: dict):
  return eval(expr, {**_BUILTINS, **ctx})  # noqa: S307

# ---------------------------------------------------------------------------
# Renderizador recursivo
# ---------------------------------------------------------------------------

class _Runner:
  """Renderizador con estado sobre una lista de tokens (compartida o slice)."""

  def __init__(self, tokens: list[tuple]):
    self.tokens = tokens
    self.pos = 0

  def _peek_kind(self) -> str | None:
    return self.tokens[self.pos][0] if self.pos < len(self.tokens) else None

  def _consume(self) -> tuple:
    t = self.tokens[self.pos]
    self.pos += 1
    return t

  def _collect_for_body(self) -> list[tuple]:
    """Recoge tokens hasta el endfor coincidente (consumido). Devuelve el cuerpo."""
    body_start = self.pos
    depth = 1
    while self.pos < len(self.tokens):
      t = self._consume()
      if t[0] == 'for':
        depth += 1
      elif t[0] == 'endfor':
        depth -= 1
        if depth == 0:
          return self.tokens[body_start: self.pos - 1]
    return self.tokens[body_start:]  # sin terminar — devolver lo que haya

  def _render_expr(self, tok: tuple, ctx: dict) -> str:
    try:
      val = _eval(tok[1], ctx)
      return '' if val is None else str(val)
    except Exception as exc:
      _collect_error(tok[1], exc)
      return f'{{ERROR:{exc}}}'

  def _render_for(self, tok: tuple, ctx: dict) -> str:
    var, list_expr = tok[1], tok[2]
    try:
      items = _eval(list_expr, ctx)
    except Exception as exc:
      _collect_error(list_expr, exc)
      items = []
    body = self._collect_for_body()
    return ''.join(_Runner(body).render({**ctx, var: _to_dotdict(item)}) for item in (items or []))

  def _render_if(self, tok: tuple, ctx: dict) -> str:
    try:
      cond = bool(_eval(tok[1], ctx))
    except Exception as exc:
      _collect_error(tok[1], exc)
      cond = False
    then_text = self.render(ctx, stop_at=frozenset({'else', 'endif'}))
    if self._peek_kind() == 'else':
      self._consume()
      else_text = self.render(ctx, stop_at=frozenset({'endif'}))
    else:
      else_text = ''
    if self._peek_kind() == 'endif':
      self._consume()
    return then_text if cond else else_text

  def render(self, ctx: dict, stop_at: frozenset = frozenset()) -> str:
    parts: list[str] = []
    while self.pos < len(self.tokens):
      kind = self._peek_kind()
      if kind in stop_at:
        break
      tok = self._consume()
      if kind == 'text':
        parts.append(tok[1])
      elif kind == 'expr':
        parts.append(self._render_expr(tok, ctx))
      elif kind == 'for':
        parts.append(self._render_for(tok, ctx))
      elif kind == 'if':
        parts.append(self._render_if(tok, ctx))
      elif kind in ('endfor', 'endif', 'else'):
        break  # terminador sin stop_at activo — ignorar
    return ''.join(parts)

# ---------------------------------------------------------------------------
# API pública
# ---------------------------------------------------------------------------

def render(template: str, context: dict) -> str:
  """
  Renderiza *template* usando *context* como espacio de nombres de evaluación.

  Los valores de *context* se envuelven automáticamente en DotDict para que
  el acceso por punto (``x.y``) funcione en las expresiones.
  """
  ctx = {k: _to_dotdict(v) for k, v in context.items()}
  return _Runner(_tokenize(template)).render(ctx)



# -*- coding: utf-8 -*-
"""
Lógica central de transformación declarativa de JSON.

El JSON transformado parte de un objeto vacío; sólo se incluyen los campos
que aparecen explícitamente en el mapeador.  El JSON original actúa como
contexto de evaluación (junto con los campos ya calculados en orden).

Formato del mapeador::

  {
    "<campo_destino>": "<campo_origen>",        # shorthand: equiv. a {"_name": "<campo_origen>"}
    "<campo_destino>": {                        # mapping completo
      "_name":   "<campo_origen>",              # campo del JSON original (por defecto: <campo_destino>)
      "_cases":  {"v1": "nv1", ..., "_default": "<por_defecto>"},
      "_value":  "cadena con {{ expr }}"
    },
    "<campo_lista>": {                          # lista-de-objetos (sin claves de control)
      "<sub_campo>": {
        "_name":   "<campo_origen>",
        "_cases":  {...},
        "_value":  "{{ expr }}"
      }
    }
  }

Un mapping dict se trata como lista-de-objetos cuando no contiene ninguna de
las claves de control ('_name', '_cases', '_value') en su nivel raíz.
"""

import ast

from . import _template as template_engine

# ---------------------------------------------------------------------------
# Constantes
# ---------------------------------------------------------------------------

_K_NAME    = "_name"
_K_CASES   = "_cases"
_K_VALUE   = "_value"
_K_DEFAULT = "_default"

_MAPPING_CONTROL_KEYS = frozenset((_K_NAME, _K_CASES, _K_VALUE))


# ---------------------------------------------------------------------------
# Helpers internos
# ---------------------------------------------------------------------------

def _parse_list(val) -> list:
  """Parsea un campo que puede ser lista Python o string repr de lista."""
  if isinstance(val, list):
    return val
  if isinstance(val, str):
    try:
      return ast.literal_eval(val)
    except Exception:
      return []
  return []


def _apply_field_to_item(item: dict, field: str, mapping: dict, ctx: dict) -> None:
  """Escribe en item[field] el valor calculado desde ctx. `_name` indica el campo fuente en ctx."""
  src = mapping.get(_K_NAME, field)
  if _K_CASES in mapping:
    original = ctx.get(src)
    new_value = mapping[_K_CASES].get(str(original), mapping[_K_CASES].get(_K_DEFAULT, original))
  elif _K_VALUE in mapping:
    new_value = template_engine.render(mapping[_K_VALUE], ctx)
  else:
    new_value = ctx.get(src)
  item[field] = new_value


def _apply_list_sub_mappings(lst: list, sub_mappings: dict, row_ctx: dict) -> list:
  """Aplica sub_mappings a cada elemento de una lista de objetos."""
  result = []
  for elem in _parse_list(lst):
    if not isinstance(elem, dict):
      result.append(elem)
      continue
    new_elem = dict(elem)
    elem_ctx = {**row_ctx, **elem}
    for sub_field, sub_mapping in sub_mappings.items():
      _apply_field_to_item(new_elem, sub_field, sub_mapping, elem_ctx)
    result.append(new_elem)
  return result


# ---------------------------------------------------------------------------
# API pública
# ---------------------------------------------------------------------------

def transform(items: list | dict, mapper: dict) -> list | dict:
  """
  Transforma cada elemento de `items` aplicando `mapper`.

  Si `items` es un objeto JSON (dict), se trata como una lista con un único
  elemento y se le asigna ``_rownum = 1``.

  El resultado de cada elemento parte de un objeto vacío; sólo se incluyen los
  campos definidos en el mapeador.  El elemento original (más ``_rownum`` y
  los campos ya calculados) actúa como contexto de evaluación.

  :param items:  Array de objetos a transformar, o un único objeto JSON.
  :param mapper: Descripción de las transformaciones (ver formato arriba).
  """
  is_list = True
  if isinstance(items, dict):
    items = [items]
    is_list = False
  elif not isinstance(items, list):
    raise TypeError("La entrada debe ser una lista o un objeto JSON")

  result = []
  for _rownum, row in enumerate(items, 1):
    new_item: dict = {}
    ctx = {**row, "_rownum": _rownum}
    for field, mapping in mapper.items():
      if isinstance(mapping, str):
        new_item[field] = ctx.get(mapping)
      elif not _MAPPING_CONTROL_KEYS.intersection(mapping):
        new_item[field] = _apply_list_sub_mappings(ctx.get(field) or [], mapping, ctx)
      else:
        _apply_field_to_item(new_item, field, mapping, ctx)
      ctx.update(new_item)
    result.append(new_item)
  if not is_list:
    return result[0]
  return result
